using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using SkinnedModel;

namespace MotionBlend {

	class Camera {
		readonly GraphicsDevice GraphicsDevice;
		public Camera(GraphicsDevice graphicsDevice) {
			GraphicsDevice = graphicsDevice;
		}
		public void Update() {
			View = Matrix.CreateLookAt(new Vector3(-1000, 300, -1000), new Vector3(0, 200, 0), Vector3.Up);
			Projection = Matrix.CreatePerspectiveFieldOfView(
				MathHelper.PiOver4, GraphicsDevice.Viewport.AspectRatio, 1, 3000);
		}
		public Matrix View;
		public Matrix Projection;
	}

	class Actor {
		Model model;
		AnimationPlayer animationPlayer;
		SkinningData skinningData;
		Camera camera;

		public Actor(Model model, Camera camera) {
			this.model = model;
			animationPlayer = new AnimationPlayer(skinningData = (SkinningData)model.Tag);
			this.camera = camera;
		}

		Matrix world = Matrix.CreateRotationY(MathHelper.Pi * 1.5f);

		public void Update() {
			animationPlayer.Update(TimeSpan.FromSeconds(1 / 60.0), world);
		}

		public void Draw() {
			Matrix[] bones = animationPlayer.GetSkinTransforms();

			foreach (ModelMesh mesh in model.Meshes) {
				foreach (Effect effect in mesh.Effects) {
					effect.Parameters["Bones"].SetValue(bones);
					effect.Parameters["View"].SetValue(camera.View);
					effect.Parameters["Projection"].SetValue(camera.Projection);
				}
				mesh.Draw();
			}
		}

		public void StartClip(string clip, TimeSpan overlapSpan, bool loop) {
			animationPlayer.StartClip(skinningData.AnimationClips[clip], overlapSpan, loop);
		}
	}

	public class MotionBlendGame : Microsoft.Xna.Framework.Game {
		readonly GraphicsDeviceManager graphics;
		SpriteBatch spriteBatch;
		Actor actor;
		Camera camera;
		KeyboardState previousState;
		KeyboardState currentState;

		public MotionBlendGame() {
			graphics = new GraphicsDeviceManager(this);
			Content.RootDirectory = "Content";
		}

		protected override void LoadContent() {
			spriteBatch = new SpriteBatch(GraphicsDevice);
			camera = new Camera(GraphicsDevice);
			actor = new Actor(Content.Load<Model>("animal"), camera);
			actor.StartClip("Sit", TimeSpan.FromSeconds(1.0), false);
			actor.StartClip("Gallop", TimeSpan.FromSeconds(1.0), true);
		}

		bool IsKeyPress(Keys Key) {
			return currentState.IsKeyDown(Key) && previousState.IsKeyUp(Key);
		}
		protected override void Update(GameTime gameTime) {
			if (GamePad.GetState(PlayerIndex.One).Buttons.Back == ButtonState.Pressed)
				this.Exit();

			previousState = currentState;
			currentState = Keyboard.GetState();

			if (IsKeyPress(Keys.Up)) {
				actor.StartClip("Stand", TimeSpan.FromSeconds(0.5), true);
			}
			if (IsKeyPress(Keys.Down)) {
				actor.StartClip("Sit", TimeSpan.FromSeconds(0.5), false);
			}
			if (IsKeyPress(Keys.Left) || IsKeyPress(Keys.Right)) {
				actor.StartClip("Gallop", TimeSpan.FromSeconds(0.5), true);
			}

			camera.Update();
			actor.Update();

			base.Update(gameTime);
		}

		protected override void Draw(GameTime gameTime) {
			graphics.GraphicsDevice.Clear(Color.LightBlue);
			actor.Draw();
			base.Draw(gameTime);
		}
	}
}
