//#define MATRIX_BLEND

#region File Description
//-----------------------------------------------------------------------------
// AnimationPlayer.cs
//
// Microsoft XNA Community Game Platform
// Copyright (C) Microsoft Corporation. All rights reserved.
//-----------------------------------------------------------------------------
#endregion

#region Using Statements
using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
#endregion

namespace SkinnedModel {
	/// <summary>
	/// The animation player is in charge of decoding bone position
	/// matrices from an animation clip.
	/// </summary>
	public class AnimationPlayer {

		class ClipPlayer {


			// Information about the currently playing animation clip.
			AnimationClip currentClipValue;
			TimeSpan currentTimeValue;
			int currentKeyframe;

			// Current animation transform matrices.
			Matrix[] boneTransforms;

			// Backlink to the bind pose and skeleton hierarchy data.
			SkinningData skinningDataValue;

			float currentBlendWeight;
			float goalBlendWeight;
			TimeSpan transitSpan;
			bool loop;

			public ClipPlayer(SkinningData skinningData) {
				if (skinningData == null)
					throw new ArgumentNullException("skinningData");

				skinningDataValue = skinningData;

				boneTransforms = new Matrix[skinningData.BindPose.Count];

			}

			public void StartClip(AnimationClip clip, bool loop) {
				if (clip == null)
					throw new ArgumentNullException("clip");

				currentClipValue = clip;
				currentTimeValue = TimeSpan.Zero;
				currentKeyframe = 0;
				this.loop = loop;

				// Initialize bone transforms to the bind pose.
				skinningDataValue.BindPose.CopyTo(boneTransforms, 0);
			}

			public void UpdateBoneTransforms(TimeSpan elapsed) {
				if (currentClipValue == null)
					throw new InvalidOperationException(
								"UpdateBoneTransforms was called before StartClip");

				bool backward = elapsed < TimeSpan.Zero;

				// Update the animation position.
				{
					currentTimeValue += elapsed;

					if (loop) {
						// If we reached the end, loop back to the start.
						while (currentTimeValue >= currentClipValue.Duration) {
							currentTimeValue -= currentClipValue.Duration;
							backward = true;
						}
					} else {
						currentTimeValue = currentClipValue.Duration;
					}
				}

				//if ((currentTimeValue < TimeSpan.Zero) || (currentTimeValue >= currentClipValue.Duration))
				//	throw new ArgumentOutOfRangeException("time");

				// If the position moved backwards, reset the keyframe index.
				if (backward) {
					currentKeyframe = 0;
					skinningDataValue.BindPose.CopyTo(boneTransforms, 0);
				}

				// Read keyframe matrices.
				IList<Keyframe> keyframes = currentClipValue.Keyframes;

				while (currentKeyframe < keyframes.Count) {
					Keyframe keyframe = keyframes[currentKeyframe];

					// Stop when we've read up to the current time position.
					if (keyframe.Time > currentTimeValue)
						break;

					// Use this keyframe.
					boneTransforms[keyframe.Bone] = keyframe.Transform;

					currentKeyframe++;
				}

				UpdateBlendWeight(elapsed);
			}

			void UpdateBlendWeight(TimeSpan elasped) {
				float d = goalBlendWeight - currentBlendWeight;

				if (d != 0) {
					if (transitSpan <= TimeSpan.Zero) {
						goalBlendWeight = currentBlendWeight;
					} else {
						double ratio = (double)elasped.Ticks / (double)transitSpan.Ticks;
						currentBlendWeight += (float)(d * ratio);
						currentBlendWeight = MathHelper.Clamp(currentBlendWeight, 0, 1);
						transitSpan -= elasped;
						if (transitSpan < TimeSpan.Zero)
							transitSpan = TimeSpan.Zero;
					}
				}
			}

			/// <summary>
			/// Gets the current bone transform matrices, relative to their parent bones.
			/// </summary>
			public Matrix[] GetBoneTransforms() {
				return boneTransforms;
			}

			/// <summary>
			/// Gets the clip currently being decoded.
			/// </summary>
			public AnimationClip CurrentClip {
				get { return currentClipValue; }
			}

			/// <summary>
			/// Gets the current play position.
			/// </summary>
			public TimeSpan CurrentTime {
				get { return currentTimeValue; }
			}

			public TimeSpan TransitSpan {
				get { return transitSpan; }
				set { transitSpan = value; }
			}
			public float CurrentBlendWeight {
				get { return currentBlendWeight; }
				set { currentBlendWeight = value; }
			}
			public float GoalBlendWeight {
				get { return goalBlendWeight; }
				set { goalBlendWeight = value; }
			}
		}

		#region Fields


		readonly ClipPlayer[] players = new ClipPlayer[2];


		// Current animation transform matrices.
		Matrix[] boneTransforms;
		Matrix[] worldTransforms;
		Matrix[] skinTransforms;


		// Backlink to the bind pose and skeleton hierarchy data.
		SkinningData skinningDataValue;


		#endregion


		/// <summary>
		/// Constructs a new animation player.
		/// </summary>
		public AnimationPlayer(SkinningData skinningData) {
			if (skinningData == null)
				throw new ArgumentNullException("skinningData");

			skinningDataValue = skinningData;

			boneTransforms = new Matrix[skinningData.BindPose.Count];
			worldTransforms = new Matrix[skinningData.BindPose.Count];
			skinTransforms = new Matrix[skinningData.BindPose.Count];

			players[0] = new ClipPlayer(skinningData);
			players[1] = new ClipPlayer(skinningData);
		}


		/// <summary>
		/// Starts decoding the specified animation clip.
		/// </summary>
		public void StartClip(AnimationClip clip, TimeSpan overlapSpan, bool loop) {
			ClipPlayer old = players[1];
			players[1] = players[0];
			players[0] = old;
			players[0].StartClip(clip, loop);

			players[0].TransitSpan = overlapSpan;
			players[0].CurrentBlendWeight = 0;
			players[0].GoalBlendWeight = 1;
			players[1].TransitSpan = overlapSpan;
			players[1].CurrentBlendWeight = 1;
			players[1].GoalBlendWeight = 0;
		}


		/// <summary>
		/// Advances the current animation position.
		/// </summary>
		public void Update(TimeSpan elapsed, Matrix rootTransform) {
			UpdateBoneTransforms(elapsed);
			UpdateWorldTransforms(rootTransform);
			UpdateSkinTransforms();
		}


		/// <summary>
		/// Helper used by the Update method to refresh the BoneTransforms data.
		/// </summary>
		public void UpdateBoneTransforms(TimeSpan elapsed) {
			players[0].UpdateBoneTransforms(elapsed);
			players[1].UpdateBoneTransforms(elapsed);

			Matrix[] boneTransforms0 = players[0].GetBoneTransforms();
			Matrix[] boneTransforms1 = players[1].GetBoneTransforms();

			float w0 = players[0].CurrentBlendWeight;
			float w1 = players[1].CurrentBlendWeight;
			float rw = 1 / (w0 + w1);
			w0 *= rw;
			w1 *= rw;

#if MATRIX_BLEND
			for (int i = 0; i < boneTransforms.Length; i++) {
				Matrix m0, m1;
				Matrix.Multiply(ref boneTransforms0[i], w0, out m0);
				Matrix.Multiply(ref boneTransforms1[i], w1, out m1);
				Matrix.Add(ref m0, ref m1, out boneTransforms[i]);
		}
#else // SRT
			for (int i = 0; i < boneTransforms.Length; i++) {
				Vector3 scale, scale0, scale1;
				Quaternion rotation, rotation0, rotation1;
				Vector3 translation, translation0, translation1;
				boneTransforms0[i].Decompose(out scale0, out rotation0, out translation0);
				boneTransforms1[i].Decompose(out scale1, out rotation1, out translation1);

				scale = scale0 * w0 + scale1 * w1;
				rotation = Quaternion.Slerp(rotation0, rotation1, w1);
				translation = translation0 * w0 + translation1 * w1;
				Matrix scaleMatrix = Matrix.CreateScale(scale);
				Matrix rotationMatrix = Matrix.CreateFromQuaternion(rotation);
				Matrix translationMatrix = Matrix.CreateTranslation(translation);
				Matrix.Multiply(ref scaleMatrix, ref rotationMatrix, out boneTransforms[i]);
				Matrix.Multiply(ref boneTransforms[i], ref translationMatrix, out boneTransforms[i]);
			}
#endif
		}


		/// <summary>
		/// Helper used by the Update method to refresh the WorldTransforms data.
		/// </summary>
		public void UpdateWorldTransforms(Matrix rootTransform) {
			// Root bone.
			worldTransforms[0] = boneTransforms[0] * rootTransform;

			// Child bones.
			for (int bone = 1; bone < worldTransforms.Length; bone++) {
				int parentBone = skinningDataValue.SkeletonHierarchy[bone];

				worldTransforms[bone] = boneTransforms[bone] *
											 worldTransforms[parentBone];
			}
		}


		/// <summary>
		/// Helper used by the Update method to refresh the SkinTransforms data.
		/// </summary>
		public void UpdateSkinTransforms() {
			for (int bone = 0; bone < skinTransforms.Length; bone++) {
				skinTransforms[bone] = skinningDataValue.InverseBindPose[bone] *
											worldTransforms[bone];
			}
		}


		/// <summary>
		/// Gets the current bone transform matrices, relative to their parent bones.
		/// </summary>
		public Matrix[] GetBoneTransforms() {
			return boneTransforms;
		}


		/// <summary>
		/// Gets the current bone transform matrices, in absolute format.
		/// </summary>
		public Matrix[] GetWorldTransforms() {
			return worldTransforms;
		}


		/// <summary>
		/// Gets the current bone transform matrices,
		/// relative to the skinning bind pose.
		/// </summary>
		public Matrix[] GetSkinTransforms() {
			return skinTransforms;
		}


	}
}
